(* ------------------------------------------------------------------- *)
(* Computes the Fractal Dimension of a given UNIX file of vectors      *)
(* ready to be called from a shell script                              *)

Off[General::spell1] (* turns off warnings *) 

(* - Utility --------------------------------------------------------- *)

lp[ l_List ] := ListPlot[ l, PlotJoined->True, PlotRange->All]

pp := PSPrint[%]

prt[ str_ ] := Print[OutputForm[str]]

pdump[ g_, fname_:"jj" ] := Display["!psfix -epsf > "<>ToString[fname], g]

rl[ f_ ] := ReadList[ f ]

(* - errorclip ------------------------------------------------------- *)
(* - given a boxcount list, it decides which part to keep.             *)
(* - The idea is that we keep the part that looks close to a line,     *)
(* - as judged by the mean squared error                               *)
(* ------------------------------------------------------------------- *)
errorclip[ bx_, minlse_:0.015, minlen_:5, verbose_:False] := 
(* ------------------------------------------------------------------- *)
Module[ { clippedbx, errMinusBottom, errMinusTop, err},

	clippedbx = bx;
	err = linerror[bx];

	While[ err> minlse && Length[clippedbx] > minlen ,
		errMinusBottom= linerror[ Drop[clippedbx, -1] ];
		errMinusTop = linerror[ Drop[clippedbx, 1] ] ; 

		If[ errMinusBottom < errMinusTop,
		    clippedbx = Drop[ clippedbx, -1], (* drop bottom *)
		    clippedbx = Drop[ clippedbx, 1] (* drop top *)
		];
		err = linerror[clippedbx];
	];

        If[verbose,
           prt[StringForm["LSE-Error = ``", err]];
           prt[StringForm["`` ``", clippedbx[[1]],
                                   clippedbx[[ Length[clippedbx] ]]]]
        ];

	If[ err > minlse ,
	    If[ verbose, Print[OutputForm["Error to big!"]]]; Return[{}],
	    clippedbx
	]

]

(* - linerror ------------------------------------------------------- *)
(* - given a list of pairs of numbers, it computes the mean squared   *)
(* - error.                                                           *)
(* ------------------------------------------------------------------ *)
linerror[bx_] := 
(* ------------------------------------------------------------------ *)
Module[ {fitfun, x, i, count, avgError},

   fitfun[x_] = Fit[ bx, {1,x}, {x}];

   count = Length[bx];
   avgError = Sum[ (fitfun[ bx[[i,1]] ] - bx[[i,2]])^2, 
			{i,1, count}];
   avgError = Chop[ Sqrt[avgError] / count ]

]

(* - boxcount2fd ---------------------------------------------------  *)
(* - computes the slope of a box-counting plot.                       *)
(* - It keeps only the part between:                                  *)
(* - ymin + tol*Abs[ymax-ymin] and ymax - tol*Abs[ymax-ymin] or,      *)
(* - if flg != 0, the part between fst and lst.                       *)
(* - Notice that bx is a list of log-values, therefore fst and lst    *)
(* - must be given as logarithm-values too.                           *)
(* - Finally it does a linear interpolation of the clipped list.      *) 
(* ------------------------------------------------------------------ *)
boxcount2fd[ bx_List, verbose_:False, stderr_:0.015, flg_, 
             fst_, lst_ ] := 
(* ------------------------------------------------------------------ *)
Module[ {ymin,ymax, yhigh, ylow, fdguess, minboxes,
         choppedbxtemp, choppedbx, pair, inbounds, fitres, err, 
         xlist, xmin, xmax, fitfun, fd, tol, avgError},

   If[flg==0,
      tol = 0.005;
      ymax = Max[ Transpose[bx][[2]] ];
      ymin = Min[ Transpose[bx][[2]] ];
      yhigh = ymax -  tol*Abs[ymax-ymin];
      ylow = ymin + tol*Abs[ymax-ymin];

      If[verbose,
         Print[OutputForm[StringForm["ylow= `` yhigh= `` tol= ``", 
                                   ylow, yhigh, tol*(ymax-ymin) ]]]
      ];
   
      (* keep only that part of bx that is within the bounds *)
      choppedbx = clipy[bx, ylow, yhigh]
   ];

   (* ------------------------------------------------------------------ *)
   (* flg==0:                                                            *)
   (* first a plot is produced, then if this does not cover more than    *)
   (* 60% of the y-range, another plot with a new linear interpolation   *)
   (* is generated.                                                      *)
   (* flg!=0:                                                            *)
   (* only one plot is generated, for the values between fst and lst     *)
 
   try = 1; diff1 = 0;
   While[try<=2,
        If[flg==0,
           len = 0;
           err = stderr;

           (* the cycle is repeated until an error smaller than err is   *)
           (* achieved, err is increased of 0.002 each time              *)
           (* ---------------------------------------------------------- *)
           While[ len < 3 && Length[choppedbx] > 3,
               choppedbxtemp = errorclip[choppedbx, err, 5, verbose];
               len = Length[choppedbxtemp];
               If[len < 3 , 
                 Print[OutputForm[
                             StringForm["too narrow strip (error ``)", err]]];
                 err = err + 0.002
               ]
           ];
           (* ---------------------------------------------------------- *)

           choppedbx = choppedbxtemp,

        (* else *)
           choppedbx = {};

           (* only the values between fst and lst are keeped             *)
           (* ---------------------------------------------------------- *)
           Do[If[(bx[[j]][[1]] >= fst) && (bx[[j]][[1]] <= lst),
                 choppedbx = Join[choppedbx,{bx[[j]]}]
              ],
             {j,1,Length[bx]}
           ];
           (* ---------------------------------------------------------- *)

           err = linerror[choppedbx]
        ];

        fitres = Fit[ choppedbx, {1,x}, {x} ];
        fitfun[x_] = Fit[ choppedbx, {1,x}, {x}];

        len = Length[choppedbx];
        diff = Abs[choppedbx[[len]][[2]] - choppedbx[[1]][[2]]];
        If[try==1, 
           fd = {{fitres[[2]][[1]],err,diff}}, 
           fd = Join[fd, {{fitres[[2]][[1]],err,diff}}]
        ];
        
        If[verbose && (flg==0),
           prt[StringForm["Length choppedbx: ``", len]];
           prt[StringForm["yhigh-ylow * 60%: ``", (yhigh - ylow)*0.6]];
           prt[StringForm["diff: ``", diff]]
        ];

        If[ verbose,
            xlist = Transpose[choppedbx][[1]];
            xmin = Min[xlist];
            xmax = Max[xlist];
            $DefaultFont = {"Times-Roman", 16};
            label1 = "log(r)";
            label2 = "log(Sq)";
            label3 = "SLOPE = "<>ToString[fitres[[2]][[1]]];
            Show[ListPlot[bx,
                          PlotRange->All, 
                          DisplayFunction->Identity],
                 Plot[fitfun[x], {x,xmin,xmax}, 
                      DisplayFunction->Identity],
                 Axes->False,
                 Frame->True,
                 FrameLabel->{label1,label2,label3,""},
                 AspectRatio->0.7,
                 DisplayFunction->$DisplayFunction
                ] 
        ];
        If[try==1 || (try==2 && (diff > diff1)),
           $DefaultFont = {"Times-Roman", 16};
           pdump[Show[ListPlot[bx,
                               PlotRange->All, 
                               DisplayFunction->Identity],
                      Plot[fitfun[x], {x,xmin,xmax}, 
                           DisplayFunction->Identity],
                      Axes->False,
                      Frame->True,
                      FrameLabel->{label1,label2,label3,""},
                      AspectRatio->0.7,
                      DisplayFunction->Identity],
                 "__post"
                ]
        ];

        If[flg==0,
           If[((yhigh - ylow)*0.6) > 
               (Abs[choppedbx[[len]][[2]] - choppedbx[[1]][[2]]]) && try==1,
              If[(yhigh - Max[choppedbx[[1]][[2]], choppedbx[[len]][[2]]]) >
                 (Min[choppedbx[[1]][[2]], choppedbx[[len]][[2]]] - ylow),
                 choppedbx = clipy[bx, 
                                Max[choppedbx[[1]][[2]],choppedbx[[len]][[2]]],
                                yhigh],
                 choppedbx = clipy[bx,
                                ylow,
                                Min[choppedbx[[1]][[2]],choppedbx[[len]][[2]]]]
              ];
              If[Length[choppedbx]>3, try = 2; diff1 = diff, try = 3],
              try = 3
           ],
           try = 3
        ]
   ];
   fd
]


(* - clipy ---------------------------------------------------------- *)
(* - clips a list of pairs of numbers, so that the y values are       *)
(* - within 'ylow' and 'yhigh'                                        *)
(* ------------------------------------------------------------------ *)
clipy[ l_, ylow_, yhigh_ ] := Select[ l,  yhigh >= #[[2]] >= ylow & ]
(* ------------------------------------------------------------------ *)

(* - vectors2boxcount ----------------------------------------------- *)
(* - given a set of vectors, it creates the boxcount plot (log-log)   *)
(* - base (1.25) by default                                           *)
(* ------------------------------------------------------------------ *)
vectors2boxcount[ s_List, q_:0, alr_, verbose_:False ] := 
(* ------------------------------------------------------------------ *)
Module[ {boxcount, minr, maxr, lowpowr, highpowr, len, base },

   {minr, maxr} = bounds[s];

   (* calculation of the r values to be used *)
   (* the list of r values must be has a log scale (100-110 points) *)
   base = 1.25;
   lowpowr = N[Ceiling[Log[minr]/Log[base] ]];
   highpowr = Floor[Log[maxr]/Log[base] ];
   While[(Abs[highpowr-lowpowr] < 99) && (base > 1.05),
         base = base-0.01;
         lowpowr = N[Ceiling[Log[minr]/Log[base] ]];
         highpowr = N[Floor[Log[maxr]/Log[base] ]];
   ];
   If[Abs[highpowr-lowpowr] > 110,
      lowpowr = lowpowr + ((Abs[highpowr-lowpowr]-110)/2);
      highpowr = highpowr - ((Abs[highpowr-lowpowr]-110))
   ];

   len = Length[s];
   Print[OutputForm["List of values representing the cell side length"]]; 
   If[verbose,
      Print[OutputForm[StringForm["min = `` --> max = `` base = `` Nval: ``", 
                                minr, maxr, base, Abs[highpowr-lowpowr]]]];
   ];
   Print[OutputForm[StringForm["min-rvalue: `` max-rvalue: ``",
                                base^lowpowr, base^highpowr]]];
   Print[OutputForm["Counting points in grid cells..."]];

   (* if there exists the file __points with the values of (r,sum pi^q)
      calculated in a previous execution (alr==1) then these values are
      used
   *)

   If[alr==0,
      boxcount = {};
      prev = If[q<=1,len,-1];
      Do[uniqueres = uniqueCount[s, base^i, q, len, prev, verbose];
         If[((q<=1) && (uniqueres<=prev)) || ((q>1) && (uniqueres>=prev)),
            If[uniqueres>0,
               AppendTo[boxcount,
                        {N[Log[base^i ]],If[q==1,uniqueres,N[Log[uniqueres]]]}]
            ];
            prev = uniqueres
         ],
         {i,lowpowr,highpowr}
      ],
   (* else *)
      boxcount = If[q==1,
                    ReadList["__points", Number, RecordLists->True],
                 (* else *)
                    N[Log[ReadList["__points", Number, RecordLists->True]]]
                 ]
   ];
   
   boxcount
]

(* - uniqueCount ---------------------------------------------------- *)
(* - it gives the sum of the pk^q, where pk is the relative           *)
(* - frequency of the points in the k-slot of the grid of size 'r'    *)
(* ------------------------------------------------------------------ *)
uniqueCount[s_List, r_, q_, num_, pr_, verbose_:False ] := 
(* ------------------------------------------------------------------ *)
Module[ {value, srt, count, tot, totend, den},

    srt = Sort[Floor[s/r]];
    den = num^q;
    (* variation for no-self pairs *)
    If[q==-1, den = num*(num-1)];
    
    If[q==0, 
       tot = Length[Union[srt]],   
    (* else *)
    (* count of duplicates and calculation of result --------------- *)
       value = srt[[1]]; count = 1; tot = 0;
       Do[If[srt[[i]] == value, 
             count++, 
          (* else *)
             If[q==1, 
                tot += (count/num) * If[count>0,Log[count/num],1],
             (* else *)
                tot += If[q!=-1,count^q,
                                (* variation for no-self pairs *)
                                If[count==1,0,count*(count-1)]]
             ]; 
             value = srt[[i]]; 
             count = 1
          ], 
          {i,2,Length[srt]}
       ];
       If[q==1, tot = N[-(tot + (count/num) * 
                                     If[count>0,Log[count/num],1])], 
       (* else *)
                tot = N[(tot + 
                         If[q!=-1,count^q,
                                  (* variation for no-self pairs *)
                                  If[count==1,0,count*(count-1)]])/den]
       ]
    ];

    (* ------------------------------------------------------------- *)
    (* If the sum is zero the defauld values: 1/den is imposed       *)
    If[tot==0,tot=N[1/den]];
    (* ------------------------------------------------------------- *)

    (* the calculated values are stored in file __res                *)
    (* ------------------------------------------------------------- *)
    If[((q<=1) && (tot<=pr)) || ((q>1) && (tot>=pr)),
       If[verbose,
          Print[OutputForm[StringForm["r: `` sum: ``",r,tot]]]; 
       ];
       If[q!=1, OutputForm[StringForm["`` ``",FortranForm[r],
                                      FortranForm[tot]]] >>> "__res", 
                OutputForm[StringForm["`` ``",FortranForm[Log[r]],
                                      FortranForm[tot]]] >>> "__res" 
       ];
    ];
    (* ------------------------------------------------------------- *)

    tot 
]

(* - bounds --------------------------------------------------------- *)
(* - functions to compute the bounds for sampling                     *)
(* - given a set of vectors, it computes the minimum and maximum      *)
(* - 'radii'                                                          *)
(* ------------------------------------------------------------------ *)
bounds[ s_List ] := 
(* ------------------------------------------------------------------ *)
Module[ { maxvalx, minvalx, dx, maxvaly, minvaly, dy, 
          count, inflation, maxr, minr},

    inflation = 4.0;
    maxvalx = Max[Transpose[s][[1]]];
    minvalx = Min[Transpose[s][[1]]];
    dx = maxvalx - minvalx;
    maxvaly = Max[Transpose[s][[2]]];
    minvaly = Min[Transpose[s][[2]]];
    dy = maxvaly - minvaly;
    count = Length[s];

    maxr = N[Sqrt[dx^2+dy^2]];
    minr = Min[ Map[mingap, Map[ usort, Transpose[s] ]] ];

    {minr/inflation, maxr*inflation}
]

(* - mingap --------------------------------------------------------- *)
(* - computes the smalles gap in a sorted list of numbers             *)
(* ------------------------------------------------------------------ *)
mingap[l_] := Min[ Abs[ Drop[ l - RotateLeft[l] , -1] ] ]
(* ------------------------------------------------------------------ *)

(* - usort ---------------------------------------------------------- *)
(* - sorts & eliminated duplicates                                    *)
(* ------------------------------------------------------------------ *)
usort[l_] := Union[Sort[l]]
(* ------------------------------------------------------------------ *)

(* ------------------------------------------------------------------ *)
(* ------------------------------------------------------------------ *)
(* - fd ------------------------------------------------------------- *)
(* - invokes the above, to compute the fractal dimension of a given   *)
(* - file of vectors                                                  *)
(* ------------------------------------------------------------------ *)
fd[ file_:"__infile", verbose_:False ] := 
(* ------------------------------------------------------------------ *)
Module[ {s, bx, q, create, fdim, initerr, already, dimlist,
         firstr, lastr, flag},

    s = ReadList[ file, Number, RecordLists->True];

    (* reading parameters                           *)
    (* -------------------------------------------- *)
    dvalue = OpenRead["__dvalue"];
    q = Read[dvalue, Number];
    create = Read[dvalue, Number];
    initerr = Read[dvalue, Number];
    already = Read[dvalue, Number];
    flag = Read[dvalue, Number];
    If[flag==1,
       firstr = Read[dvalue, Number];
       lastr = Read[dvalue, Number],
    ];

    Close[dvalue];
    (* -------------------------------------------- *)
    
    dimlist = Length[s[[1]]];
    bx = vectors2boxcount[s,q,already,verbose];
    fdim = boxcount2fd[bx, verbose, initerr, flag, firstr, lastr];
    Print[OutputForm[StringForm["Q value: ``",q]]];
    Print[OutputForm["Fractal dimension is:"]]; 
    If[Length[fdim]==2,
       If[fdim[[1]][[3]] > fdim[[2]][[3]],
          OutputForm[fdim[[1]][[2]]] >> __fd;
          Which[ q <= 1, -fdim[[1]][[1]], 
                 q > 1, N[fdim[[1]][[1]]/(q-1)]] >>> __fd;
          OutputForm[fdim[[2]][[2]]] >>> __fd;
          Which[ q <= 1, -fdim[[2]][[1]], 
                 q > 1, N[fdim[[2]][[1]]/(q-1)]] >>> __fd,
          OutputForm[fdim[[2]][[2]]] >> __fd;
          Which[ q <= 1, -fdim[[2]][[1]], 
                 q > 1, N[fdim[[2]][[1]]/(q-1)]] >>> __fd;
          OutputForm[fdim[[1]][[2]]] >>> __fd;
          Which[ q <= 1, -fdim[[1]][[1]], 
                 q > 1, N[fdim[[1]][[1]]/(q-1)]] >>> __fd
       ],
       OutputForm[fdim[[1]][[2]]] >> __fd;
       Which[ q <= 1, -fdim[[1]][[1]], 
              q > 1, N[fdim[[1]][[1]]/(q-1)]] >>> __fd
    ]; 
    If[(create==1) && dimlist < 3, 
       pdump [Graphics[ListPlot[s,PlotRange->All,DisplayFunction->Identity],
                       DisplayFunction->Identity], "__plot"]
    ];

    If[Length[fdim]==2,
       If[fdim[[1]][[3]] > fdim[[2]][[3]],
          OutputForm[fdim[[1]][[2]]];
          Which[ q <= 1, -fdim[[1]][[1]], q > 1, N[fdim[[1]][[1]]/(q-1)]],
          OutputForm[fdim[[2]][[2]]];
          Which[ q <= 1, -fdim[[2]][[1]], q > 1, N[fdim[[2]][[1]]/(q-1)]]
       ],
       OutputForm[fdim[[1]][[2]]];
       Which[ q <= 1, -fdim[[1]][[1]], 
              q > 1, N[fdim[[1]][[1]]/(q-1)]]
    ]
]

(* the command that does the whole job *)
fd["__infile", True]
